# scheme.py
#
# Challenge: Can you implement a scheme interpreter in Python that's
# capable of executing the following procedure?

# A procedure definition for:
#
#   (define fact
#      (lambda (n) (if (= n 1)
#                   1
#                   (* n (fact (- n 1))))))
#
# It's represented in Python using the following tuple:

fact = ('define', 'fact',
        ('lambda', ('n',), ('if', ('=', 'n', 1),
                            1,
                            ('*', 'n', ('fact', ('-', 'n', 1))))))


def define(label, value):
    global env
    env[label]=value


env = {
    "two": 2,
    "+": lambda x, y: x+y,
    "*": lambda x, y: x*y,
    "-": lambda x, y: x-y,
    "/": lambda x, y: x/y,
    "<": lambda x, y: x<y,
    ">": lambda x, y: x>y,
    "=": lambda x, y: x == y,
}


def substitute(exp, name, value):
    if exp == name:
        return value
    elif isinstance(exp, tuple):
        return tuple(substitute(e, name, value) for e in exp)
    else:
        return exp


def make_procedure(argnames, lexp):
    """Returns a procedure"""
    def call(*values):
        assert len(values) == len(argnames), "Wrong, # args"
        exp = lexp
        # substitute the names
        for name, val in zip(argnames, values):
            exp = substitute(exp, name, val)
        return seval(exp)
    return call


# You will define the following procedure for evaluating an expression
def seval(sexp):
    """Scheme eval"""
    global env
    if isinstance(sexp, (int, bool, float)):
        return sexp
    elif isinstance(sexp, str):
        return env[sexp]
    elif isinstance(sexp, tuple):
        if sexp[0] == "define":
            #######
        elif sexp[0] == "if":
            #######
        elif sexp[0] == "lambda":
            #######
        else:
            evaluated_exp = [seval(e) for e in sexp]
            return evaluated_exp[0](*evaluated_exp[1:])


# In writing seval, you are ONLY allowed to use the rules of Scheme
# evaluation that you currently know about.  So far, this includes the
# substitution model and the notion of special forms.

# Some basic tests
assert seval(42) == 42
assert seval("two") == 2

assert seval(('+', 2, 3)) == 5
assert seval(('+', ("*", 33, 44), 3)) is not None
seval(('define', 'n', 5))
assert seval('n') == 5

assert seval(("if", (">", 1, 2), 10, 12)) == 12


assert seval(("lambda", ("x", "y"), ("+", "x", "y"))) is not None, "Lambda isn't working"

# Now the ultimate test--can you run your procedure?
seval(fact)

assert seval(('fact', 'n')) == 120
